'''
Created on 13 Aug 2016
@author: rizarse
'''

class Message(object):

    def __init__(self, topic, channel, payload, index = 0, timestamp = 0):
        
        assert topic is not None, 'Topic is required'
        assert channel is not None, 'Channel is required' 
        assert payload is not None, 'Topic is required'
        
        assert isinstance(topic, str),' Topic must be type of str'
        assert isinstance(channel, int),' Topic must be type of int'
        assert isinstance(payload, bytes),' Topic must be type of bytes'
        
        self.__topic = topic;
        self.__channel = channel;
        self.__body = payload;
        self.__index = index;
        self.__timestamp = timestamp;
    
    def topic(self):
        return self.__topic;
    
    def channel(self):
        return self.__channel;
    
    def payload(self):
        return self.__body;
    
    def index(self):
        return self.__index;
    
    def timestamp(self):
        return self.__timestamp;


'''
Created on 11 Aug 2016
@author: rizarse
'''

import logging
import threading

from os.path import expanduser

from kafka.consumer.group import KafkaConsumer
from magistral.client.Configs import Configs
from magistral.client.MagistralException import MagistralException
from kafka.structs import TopicPartition
from magistral.Message import Message

class GroupConsumer(threading.Thread):
        
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
     
    def __init__(self, threadId, name, sKey, bootstrapServers, groupId, permissions, cipher = None, uid = None):
        threading.Thread.__init__(self)
        self.threadId = threadId
        self.name = name
        
        self.group = groupId;
        self.subKey = sKey;
        
        self.cipher = None if cipher is None else cipher;
                
        configs = Configs.consumerConfigs();
        configs["bootstrap_servers"] = bootstrapServers;
        configs["group_id"] = groupId;
        configs['enable_auto_commit'] = False;
        
        self.__isAlive = True
        
        home = expanduser("~")
        
        if uid == None:
            self.__consumer = KafkaConsumer(
                bootstrap_servers = bootstrapServers,
                check_crcs = False,
                exclude_internal_topics = True,
                session_timeout_ms = 20000,
                fetch_min_bytes = 128,
                fetch_max_wait_ms = 256,           
                enable_auto_commit = False,
                max_in_flight_requests_per_connection = 10,                
                group_id = groupId);
        else:
            self.__consumer = KafkaConsumer(
                bootstrap_servers = bootstrapServers,
                check_crcs = False,
                exclude_internal_topics = True,
                session_timeout_ms = 20000,
                fetch_min_bytes = 128,
                fetch_max_wait_ms = 256,           
                enable_auto_commit = False,
                max_in_flight_requests_per_connection = 10,
                security_protocol = 'SSL',
                ssl_check_hostname = False,
                ssl_keyfile = home + '/magistral/' + uid + '/key.pem',
                ssl_cafile = home + '/magistral/' + uid + '/ca.pem',
                ssl_certfile = home + '/magistral/' + uid + '/certificate.pem',
                group_id = groupId);
        
        self.permissions = permissions;
        self.map = {}
        
        self.__offsets = {}
        
        
    def recordsTotally(self, data):
        size = 0;
        for val in data.values(): 
            if len(val) > 0: size = size + len(val);
                               
        return size;
    
    def consumerRecord2Message(self, record):                    
        payload = record[6]
                                        
        if self.cipher is not None:
            try:
                payload = self.cipher.decrypt(payload)
            except:
                pass
                                        
        msg = Message(record[0], record[1], payload, record[2], record[3])
        return msg

    def run(self):
        
        threadLock.acquire()
        
        while self.__isAlive:
            try:

                data = self.__consumer.poll(512);
                for values in data.values():
                                         
                    for value in values:
                        msg = self.consumerRecord2Message(value);
                        listener = self.map[msg.topic()][msg.channel()];
                        if listener is not None: listener(msg);
                        
                self.__consumer.commit_async(); 
                    
            except:
                pass
              
        threadLock.release()

#   ////////////////////////////////////////////////////////////////////////////////////    

    def subscribe(self, topic, channel = -1, listener = None, callback = None):
        
        assert channel is not None and isinstance(channel, int), "Channel expected as int argument"        
        if (channel < -1): channel = -1;
                
        etopic = self.subKey + "." + topic;
                           
        self.logger.debug("Subscribe -> %s : %s | key = %s", topic, channel, self.subKey);
        
        if (self.permissions == None or len(self.permissions) == 0): 
            raise MagistralException("User has no permissions for topic [" + topic + "].");
        
        self.fch = [];
        
        for meta in self.permissions:             
            if (meta.topic() != topic): continue;  
            
            if channel == -1:
                self.fch = meta.channels();
            elif channel in meta.channels():                          
                self.fch = [ channel ];  
        
        if (len(self.fch) == 0): 
            npgex = "No permissions for topic [" + topic + "] granted";
            self.logger.error(npgex);                                
            raise MagistralException(npgex);
        
        if (self.map == None or etopic not in self.map): 
            self.map[etopic] = {}
        
#         // Assign Topic-partition pairs to listen
        
        tpas = [];
        for ch in self.fch:            
            tpas.append(TopicPartition(etopic, ch));
            if (listener is not None): self.map[etopic][ch] = listener;        
        
        self.__consumer.assign(tpas);       
                
        if callback is not None: 
            callback(self.__consumer.assignment());
            
        return self.__consumer.assignment();
        
        
    def unsubscribe(self, topic):
        self.consumer.assign([]);
        self.map.remove(topic);        

    def close(self):
        self.__isAlive = False
        self.__consumer.pause()
        self.__consumer.close()
        
    logging.getLogger('kafka.conn').setLevel(logging.FATAL)
    logging.getLogger('kafka.cluster').setLevel(logging.FATAL)
    logging.getLogger('kafka.consumer.group').setLevel(logging.INFO)    
    logging.getLogger('kafka.consumer.fetcher').setLevel(logging.INFO)
    logging.getLogger('kafka.coordinator.consumer').setLevel(logging.INFO)
    logging.getLogger('kafka.producer.record_accumulator').setLevel(logging.INFO)
    
threadLock = threading.Lock()    


'''
Created on 10 Aug 2016

@author: rizarse
'''
from enum import Enum

class MagistralExceptionType(Enum):
    
    GeneralError = "MagistralException"
    
    ConnectionPointsError = "Error while receiving connection points"    
    MqttConnectionError = "Error connecting to MQTT"
    RestError = "REST service execution error"
    
    ConversionError = "Conversion error"
    
    HistoryInvocationError = ""
    FetchTopicsError = "Error when fetching topic information"
    
    PermissionFetchError = "Error when fetching user permissions"
    PermissionGrantError = "Error when granting user permissions"
    PermissionRevokationError = "Error when revoking user permissions"
    
    InvalidPubKey = "Invalid publish key"
    InvalidSubKey = "Invalid subscription key"
    InvalidSecretKey = "Invalid secret key"


'''
Created on 9 Aug 2016

@author: rizarse
'''

import json
import logging
from magistral.client.perm.PermMeta import PermMeta
from magistral.client.Configs import Configs

class JsonConverter(object):
    
    logger = logging.getLogger(__name__)
    
    @staticmethod
    def connectionSettings(_json):
        parsedJson = json.loads(_json)
        
        settings = {}
        pub = {}
        sub = {}
        
        token = ""
        
        for entry in parsedJson:
            
            if entry["producer"] != None: 
                pp = Configs.producerConfigs();
                pp["bootstrap_servers"] = entry["producer"];
                
                if "plain" in pub:
                    pub["plain"].append(pp);
                else:
                    pub["plain"] = [pp]
                
            if entry["producer-ssl"] != None: 
                pp = Configs.producerConfigs();
                pp["bootstrap_servers"] = entry["producer-ssl"];               
                
                if "ssl" in pub:
                    pub["ssl"].append(pp);
                else:
                    pub["ssl"] = [pp]
                
            if entry["consumer"] != None: 
                sp = {}
                sp["bootstrap_servers"] = entry["consumer"];
                
                if "plain" in sub:
                    sub["plain"].append(sp);
                else:
                    sub["plain"] = [sp]
                
            if entry["consumer-ssl"] != None: 
                sp = {}
                sp["bootstrap_servers"] = entry["consumer-ssl"];
                                
                if "ssl" in sub:
                    sub["ssl"].append(sp);
                else:
                    sub["ssl"] = [sp]
            
            if entry["token"] != None:
                token = entry["token"] 
                settings["meta"] = { "token" : entry["token"] }
          
            if entry["ts"] != None:
                settings["ts"] = entry["ts"]
                
            if entry["ks"] != None:
                settings["ks"] = entry["ks"]
        
        settings["pub"] = pub;
        settings["sub"] = sub;        
        
        JsonConverter.logger.debug("Token = %s", token)
        return settings 
    
    @staticmethod
    def __convertPerm(entry):
        permissions = {}

        read = entry["read"];
        write = entry["write"];
        channels = entry["channels"];
                
        if isinstance(channels, list):
            for ch in channels:
                permissions[int(ch)] = (read, write)                     
        else:
            permissions[int(channels)] = (read, write)
        
        return permissions;
    
    @staticmethod
    def userPermissions(_json):
        if (_json == None or _json == 'null'): return [];
                
        parsedJson = json.loads(_json)
        
        permissions = [];
                
        ps = parsedJson["permission"]
        
        if isinstance(ps, list):
            for p in ps: 
                permissions.append(PermMeta(p["topic"], JsonConverter.__convertPerm(p)))
        else:            
            permissions.append(PermMeta(ps["topic"], JsonConverter.__convertPerm(ps)))
                        
        return permissions


'''
Created on 11 Aug 2016
@author: rizarse
'''

class SubMeta(object):
    
    def __init__(self, group, topic, channel, endPoints = None):
        
        assert topic is not None, 'Topic name required'
        assert channel is not None, 'Channel No required'
        
        assert isinstance(topic, str), 'Topic name must be type of str'
        assert isinstance(channel, int), 'Channel must be type of str'
        
        self.__group = group
        self.__topic = topic        
        self.__endPoints = endPoints
        self.__channel = channel
    
    def group(self):
        return self.__group
    
    def topic(self):
        return self.__topic
    
    def channel(self):
        return self.__channel
    
    def endPoints(self):
        return self.__endPoints


'''
Created on 13 Aug 2016
@author: rizarse
'''

from abc import abstractmethod
import time

class IHistory(object):
    
    @abstractmethod
    def history(self, topic, channel, count, start = 0, callback = None): pass

    @abstractmethod
    def historyIn(self, topic, channel, start = 0, end = int(round(time.time() * 1000)), callback = None): pass


'''
Created on 24 Aug 2016
@author: rizarse
'''

import binascii
from Crypto.Cipher import AES
import base64

class AESCipher:
    '''
    PyCrypto AES using ECB mode implementation in Python 3.3.  
    This uses very basic 0x00 padding, I would recommend PKCS5/7.
    '''

    def __init__(self, key):
        '''
        The constructor takes in a PLAINTEXT string as the key and converts it
        to a byte string to work with throughout the class.
        '''
        # convert key to a plaintext byte string to work with it
        self.key = bytes(key, encoding='utf-8')
        self.BLOCK_SIZE = 16
        
    def __pad(self, raw):
        '''
        This right pads the raw text with 0x00 to force the text to be a
        multiple of 16.  This is how the CFX_ENCRYPT_AES tag does the padding.
        
        @param raw: String of clear text to pad
        @return: byte string of clear text with padding
        '''
        if (len(raw) % self.BLOCK_SIZE == 0):
            return raw
        
        padding_required = self.BLOCK_SIZE - (len(raw) % self.BLOCK_SIZE)
        padChar = b'\x00'
        data = raw + padding_required * padChar
        return data
    
    def __unpad(self, s):
        '''
        This strips all of the 0x00 from the string passed in. 
        
        @param s: the byte string to unpad
        @return: unpadded byte string
        '''
        s = s.rstrip(b'\x00')
        return s
    
    def encrypt(self, raw):
        '''
        Takes in bytes and encrypts it.
        
        @param raw: a string of clear text
        @return: a string of encrypted ciphertext
        '''
        assert raw is not None and len(raw) > 0, 'input text cannot be null or empty set'
        assert isinstance(raw, bytes)

        # padding put on before sent for encryption
        raw = self.__pad(raw)
        cipher = AES.AESCipher(self.key[:32], AES.MODE_ECB)
        ciphertext = cipher.encrypt(raw)
        return base64.b64encode(ciphertext).decode("utf8")
#         return  binascii.hexlify(bytearray(ciphertext)).decode('utf-8')
    
    def decrypt(self, enc):
        '''
        Takes in a b64-string of ciphertext and decrypts it.
        
        @param enc: encrypted string of ciphertext
        @return: decrypted bytes
        '''                
        assert enc is not None and len(enc) > 0, 'input text cannot be null or empty set'
        
        enc = base64.b64decode(enc)
        cipher = AES.AESCipher(self.key[:32], AES.MODE_ECB)
        enc = self.__unpad(cipher.decrypt(enc))
        return enc
        


'''
Created on 16 Sep 2016
@author: rizarse
'''

import jks, textwrap, base64
from os.path import expanduser
import os.path

from os import makedirs

class JksHandler(object):

    def __init__(self, params):
        pass
    
    @staticmethod
    def writePkAndCerts(ks):
        
        uid = None
        
        home = expanduser("~")
        
        for alias, pk in ks.private_keys.items(): 
            
            uid = alias
            
            if pk.algorithm_oid == jks.util.RSA_ENCRYPTION_OID:
                
                if os.path.exists(home + '/magistral/' + uid) == False:
                    makedirs(home + '/magistral/' + uid)
                
                key = home + '/magistral/' + uid + '/key.pem'
                
                if os.path.exists(key): os.remove(key) 
                                
                with open(key, 'wb') as f:
                    f.seek(0)
                    f.write(bytearray(b"-----BEGIN RSA PRIVATE KEY-----\r\n"))
                    f.write(bytes("\r\n".join(textwrap.wrap(base64.b64encode(pk.pkey).decode('ascii'), 64)), 'utf-8'))
                    f.write(bytearray(b"\r\n-----END RSA PRIVATE KEY-----"))                    
                    f.close()   
             
            counter = 0;
            
            cert = home + '/magistral/' + uid + '/certificate.pem'
            if os.path.exists(cert): os.remove(cert)
            
            with open(cert, 'wb') as f:
                f.seek(0)
                   
                for c in pk.cert_chain:               
                    f.write(bytearray(b"-----BEGIN CERTIFICATE-----\r\n"))
                    f.write(bytes("\r\n".join(textwrap.wrap(base64.b64encode(c[1]).decode('ascii'), 64)), 'utf-8'))
                    f.write(bytearray(b"\r\n-----END CERTIFICATE-----\r\n"))   
                    counter = counter + 1
                    if (counter == 2): break                 
                
                f.close()
            
            ca = home + '/magistral/' + uid + '/ca.pem'
            if os.path.exists(ca): os.remove(ca)
            
            with open(ca, 'wb') as f:         
                for alias, c in ks.certs.items():    
                    f.write(bytearray(b"-----BEGIN CERTIFICATE-----\r\n"))
                    f.write(bytes("\r\n".join(textwrap.wrap(base64.b64encode(c.cert).decode('ascii'), 64)), 'utf-8'))
                    f.write(bytearray(b"\r\n-----END CERTIFICATE-----\r\n"))                    
                    
                f.close()
            
        return uid
                 
    
    @staticmethod
    def printJks(ks):
        
        def print_pem(der_bytes, type):
            print("-----BEGIN %s-----" % type)
            print("\r\n".join(textwrap.wrap(base64.b64encode(der_bytes).decode('ascii'), 64)))
            print("-----END %s-----" % type)

    
        for alias, pk in ks.private_keys.items():
            print("Private key: %s" % pk.alias)
            if pk.algorithm_oid == jks.util.RSA_ENCRYPTION_OID:
                print_pem(pk.pkey, "RSA PRIVATE KEY")
            else:
                print_pem(pk.pkey_pkcs8, "PRIVATE KEY")
        
            for c in pk.cert_chain:
                print_pem(c[1], "CERTIFICATE")
            print()
        
        for alias, c in ks.certs.items():
            print("Certificate: %s" % c.alias)
            print_pem(c.cert, "CERTIFICATE")
            print()
        
        for alias, sk in ks.secret_keys.items():
            print("Secret key: %s" % sk.alias)
            print("  Algorithm: %s" % sk.algorithm)
            print("  Key size: %d bits" % sk.key_size)
            print("  Key: %s" % "".join("{:02x}".format(b) for b in bytearray(sk.key)))
            print()    



'''
Created on 9 Aug 2016

@author: rizarse
'''

import logging

class Configs:
    logging.basicConfig(level=logging.INFO,
                        format=('%(filename)s: '    
                                '%(levelname)s: '
                                '%(funcName)s(): '
                                '%(lineno)d:\t'
                                '%(message)s')
                        )
        
    @staticmethod
    def producerConfigs():
        producer = {}
        producer["producer_type"] = "async";
        producer["acks"] = "1"
        producer["retries"] = 5
        producer["batch_size"] = 32768
        producer["linger_ms"] = 5
        producer["request_timeout_ms"] = 60000        
        producer["buffer_memory"] = 33554432
        producer["key_serializer"] = None
        producer["value_serializer"] = None
        producer["compression_type"] = "gzip"
        return producer;
     
    @staticmethod   
    def consumerConfigs():
        consumer = {}
        consumer["enable_auto_commit"] = False;
        consumer["check_crcs"] = False;
        consumer["exclude_internal_topics"] = True;
        consumer["session_timeout_ms"] = 30000       
        consumer["fetch_min_bytes"] = 8
        consumer["fetch_wait_max_ms"] = 256
        consumer["max_in_flight_requests_per_connection"] = 4
        consumer["buffer_memory"] = 33554432
        consumer["key_serializer"] = None
        consumer["value_serializer"] = None
        return consumer;
    
    


'''
Created on 10 Aug 2016
@author: rizarse
'''

class PermMeta():
    
    def __init__(self, topic, perms):
        assert topic is not None, 'Topic name required'
        assert isinstance(topic, str), 'Topic name must be type of str'
        
        self._topic = topic
        self._perms = perms
        
    def topic(self):
        return self._topic
    
    def channels(self):
        return self._perms.keys()
    
    def readable(self, channel):
        if (self._perms == None or self._perms.has_key(channel) == False): 
            return False
        else:
            return self._perms[channel][0]
    
    def writable(self, channel):
        if (self._perms == None or self._perms.has_key(channel) == False): 
            return False
        else:
            return self._perms[channel][1]


'''
Created on 13 Aug 2016
@author: rizarse
'''

from kafka.consumer.group import KafkaConsumer
from kafka.structs import TopicPartition
from magistral.Message import Message
from magistral.client.MagistralException import MagistralException

class MagistralConsumer(object):
    
    __HISTORY_DATA_FETCH_SIZE_LIMIT = 10000;

    def __init__(self, pubKey, subKey, secretKey, bootstrap, cipher = None):
        self.__pubKey = pubKey
        self.__subKey = subKey
        self.__secretKey = secretKey
        
        self.__bootstrap = bootstrap
        if cipher is not None: self.__cipher = cipher
    
    def history(self, topic, channel, records):
        
        messages = []
        
        consumer = KafkaConsumer(bootstrap_servers = self.__bootstrap,
            enable_auto_commit = False, session_timeout_ms = 30000, fetch_min_bytes = 32, max_partition_fetch_bytes = 65536);
        
        if (records > self.__HISTORY_DATA_FETCH_SIZE_LIMIT): records = self.__HISTORY_DATA_FETCH_SIZE_LIMIT;
        
        kfkTopic = self.__subKey + "." + topic;
        x = TopicPartition(kfkTopic, channel);
        
        consumer.assign([x]);
        consumer.seek_to_end();        
        last = consumer.position(x);
        
        pos = last - records if last > records else 0;
        consumer.seek(x, pos);
        
        data = consumer.poll(256);   
        
        endIsNotReached = True;
        while endIsNotReached:
            
            if len(data.values()) == 0:
                return messages;
            
            records = list(data.values())
            
            for record in records[0]:
                index = record[2];
                if index >= last - 1: endIsNotReached = False;
                
                message = Message(record[0], record[1], record[6], index, record[3]);
                messages.append(message);
            
            if endIsNotReached == False: 
                consumer.close();
                return messages;
            
            pos = pos + len(messages)
            consumer.seek(x, pos);
            data = consumer.poll(256);
            
        consumer.close();
        
        return messages;
    
    def historyForTimePeriod(self, topic, channel, start, end, limit = -1):
        
        out = []        
        
        try:
            kfkTopic = self.__subKey + "." + topic;
            x = TopicPartition(kfkTopic, channel);
                        
            consumer = KafkaConsumer(bootstrap_servers = self.__bootstrap);
            consumer.assign([x]);
        
            consumer.config['enable_auto_commit'] = False;
            consumer.config['session_timeout_ms'] = 30000;
            consumer.config['fetch_min_bytes'] = 32;
            consumer.config['max_partition_fetch_bytes'] = 65536;
            
            consumer.seek_to_end();        
            last = consumer.position(x);
        
            position = last - 1000;
            
            found = False;
            while found == False:
                consumer.seek(x, position);
                data = consumer.poll(500);
                 
                if x not in data.keys() or len(data[x]) == 0: break;
                
                record = data[x][0];
                timestamp = record[3];
 
                if timestamp < start: 
                    found = True;
                    break;
                 
                position = position - 1000;
             
            consumer.close();
                       
            c = KafkaConsumer(bootstrap_servers = self.__bootstrap);            
            c.assign([x]);
             
            c.config['fetch_min_bytes'] = 32;
            c.config['max_partition_fetch_bytes'] = 65536;
                       
            c.seek(x, position);                        
            data = c.poll(256);
            
            while (x in data.keys() and len(data[x]) > 0):
                
                for record in data[x] :
                    timestamp = record[3];
                    if timestamp < start: continue;
                    
                    index = record[2];
                    
                    if timestamp > end or index >= last - 1:  
                        c.close();                 
                        return out;
                                    
                    message = Message(record[0], record[1], record[6], index, timestamp);
                    out.append(message); 
                    
                    if limit is not None and limit > 0 and len(out) >= limit:
                        c.close();                 
                        return out;                 
                    
                c.seek(x, position + len(data[x]));                        
                data = c.poll(256);
            
            return out;
        
        except:
            
            raise MagistralException("Exception during history invocation occurred");



'''
Created on 13 Aug 2016
@author: rizarse
'''

import time

class PubMeta(object):

    def __init__(self, topic, channel, timestamp = int(round(time.time() * 1000))):
        self.__topic = topic
        self.__channel = channel
        self.__timestamp = timestamp
        
    def topic(self):
        return self.__topic
    
    def channel(self):
        return self.__channel
        
    def timestamp(self):
        return self.__timestamp


'''
Created on 9 Aug 2016
@author: rizarse
'''

from abc import abstractmethod


class IAccessControl :
    
    @abstractmethod
    def permissions(self, topic = None, callback = None): pass
    
    @abstractmethod
    def grant(self, user, topic, read, write, ttl = 0, channel = -1, callback = None): pass
    
    @abstractmethod
    def revoke(self, user, topic, channel = -1, callback = None): pass


'''
Created on 9 Aug 2016

@author: rizarse
'''

import requests
import logging
from magistral.client.MagistralException import MagistralException
# from magistral.client.MagistralExceptionType import MagistralExceptionType

class RestApiManager:        
   
    logging.basicConfig(level = logging.DEBUG);    
    logger = logging.getLogger(__name__);

    @staticmethod
    def get(path, params = None, user = None, password = None, callback = None):
        
        log = RestApiManager.logger;
        log.debug("HTTP GET: %s %s %s %s", path, params, user, password)
        
        r = None;
        
        if params == None:
            if user == None or password == None:
                r = requests.get(path);
            else:
                r = requests.get(path, auth=(user, password));
        else:
            if user == None or password == None:
                r = requests.get(path, params=params);
            else:
                s = requests.Session()
                s.auth = (user, password)
                s.params = params
                
                r = s.get(path);
        
        if r != None:
            if (r.ok):
                log.debug("JSON: %s", r.text);
                if (callback != None): callback(r.text, None);
                return r.text;                
            else:            
                log.error("ERROR: [%d] %s", r.status_code, r.text, r.content);
                if (callback != None): callback(r.text, MagistralException(r.text));
                return None;            
        else:
            log.error("ERROR: %S", r.text);
            if (callback != None): callback(None, MagistralException(r.text)); 
            return None;
        
    @staticmethod
    def put(path, params = None, user = None, password = None, callback = None):
        
        log = RestApiManager.logger;
        log.debug("HTTP PUT: %s %s %s %s", path, params, user, password)
        
        r = None;
        
        if params == None:
            if user == None or password == None:
                r = requests.put(path);
            else:
                r = requests.put(path, auth=(user, password));
        else:
            if user == None or password == None:
                r = requests.put(path, params=params);
            else:
                s = requests.Session()
                s.auth = (user, password)
                s.params = params
                
                r = s.put(path);
        
        if r != None:
            if (r.ok):
                log.debug("JSON: %s", r.text);
                if (callback != None): callback(r.text, None);
                return r.text;              
            else:            
                log.error("ERROR: [%d] %s", r.status_code, r.text, r.content);
                if (callback != None): callback(r.text, MagistralException(r.text));   
                return None;         
        else:
            log.error("ERROR: %S", r.text);
            if (callback != None): callback(None, MagistralException(r.text));
            return None;
            
    @staticmethod
    def delete(path, params = None, user = None, password = None, callback = None):
        
        log = RestApiManager.logger;
        log.debug("HTTP PUT: %s %s %s %s", path, params, user, password)
        
        r = None;
        
        if params == None:
            if user == None or password == None:
                r = requests.delete(path);
            else:
                r = requests.delete(path, auth=(user, password));
        else:
            if user == None or password == None:
                r = requests.delete(path, params=params);
            else:
                s = requests.Session()
                s.auth = (user, password)
                s.params = params
                
                r = s.delete(path);
        
        if r != None:
            if (r.ok):
                log.debug("JSON: %s", r.text);
                if (callback != None): callback(r.text, None);  
                return r.text;              
            else:            
                log.error("ERROR: [%d] %s", r.status_code, r.text, r.content);
                if (callback != None): callback(r.text, MagistralException(r.text));   
                return None;         
        else:
            log.error("ERROR: %S", r.text);
            if (callback != None): callback(None, MagistralException(r.text)); 
            return None;
        


'''
Created on 12 Aug 2016
@author: rizarse
'''

class TopicMeta(object):

    def __init__(self, topic, channels):        
        assert topic is not None, 'Topic name required'
        assert isinstance(topic, str), 'Topic name must be type of str'
        
        self.__topicName = topic
        self.__channels = channels

    def topic(self):
        return self.__topicName
    
    def channels(self):
        return self.__channels


'''
Created on 10 Aug 2016
@author: rizarse
'''

class MagistralException(Exception):
       
    def __init__(self, value):
        self.value = value
        
    def __str__(self):
        return repr(self.value)


from distutils.core import setup
from setuptools import setup, find_packages

setup(
  name = 'magistral',
  packages = find_packages(),
  install_requires = ['pyjks>=0.5.1', 'requests>=2.11.1', 'paho--mqtt>=1.1', 'kafka>=1.3.1', 'pycrypto>=1.1.4', 'xxhash>=0.6.1', 'lz4>=0.8.1', 'lz4tools>=1.3.1.2'],
  version = '0.5.3',
  description = 'Python SDK for Magistral',
  author = 'Magistral.IO team',
  author_email = 'admin@magistral.io',
  url = 'https://github.com/magistral-io/MagistralPython',
  download_url = 'https://github.com/magistral-io/MagistralPython/tarball/0.5',
  keywords = [ 'magistral', 'sdk', 'messaging' ],
  classifiers = [],
)


'''
Created on 9 Aug 2016

@author: rizarse
'''

from abc import abstractmethod

class IMagistral :
    
    @abstractmethod
    def subscribe(self, topic, group = "default", channel = -1, listener = None, callback = None): pass
    
    @abstractmethod
    def unsubscribe(self, topic, channel = -1, callback = None): pass
    
    @abstractmethod
    def publish(self, topic, msg, channel = -1, callback = None): pass
    
    @abstractmethod
    def topics(self, callback = None): pass
    
    @abstractmethod
    def topic(self, topic, callback = None): pass
    
    @abstractmethod
    def close(self): pass
        


'''
Created on 9 Aug 2016
@author: rizarse
'''

import re
import sys
import jks
import time
import base64
import logging
import paho.mqtt.client as mqtt
from magistral.client.util.JksHandler import JksHandler

from os.path import expanduser, os

from magistral.client.IAccessControl import IAccessControl
from magistral.client.IMagistral import IMagistral
from magistral.client.util.RestApiManager import RestApiManager
from magistral.client.util.JsonConverter import JsonConverter
from magistral.client.sub.GroupConsumer import GroupConsumer
from magistral.client.sub.SubMeta import SubMeta
from magistral.client.topics.TopicMeta import TopicMeta
from magistral.client.MagistralException import MagistralException

from kafka.producer.kafka import KafkaProducer
from kafka.producer.future import RecordMetadata
from magistral.client.pub.PubMeta import PubMeta
from magistral.client.IHistory import IHistory

from magistral.client.sub.MagistralConsumer import MagistralConsumer
from magistral.client.util.aes import AESCipher
import shutil


class Magistral(IMagistral, IAccessControl, IHistory):

    logger = logging.getLogger(__name__);
    logger.setLevel(logging.INFO)
    
    __host = "app.magistral.io"
    
    def __init__(self, pubKey, subKey, secretKey, cipher = None):
        
        assert pubKey is not None and subKey is not None and secretKey is not None, 'Publish, Subscribe and Secret key must be specified' 
        assert isinstance(pubKey, str) and isinstance(subKey, str) and isinstance(secretKey, str), 'Publish, Subscribe and Secret key must be type of str'
        
        pk_regex = re.compile('^pub-[a-f0-9]{8}-?[a-f0-9]{4}-?4[a-f0-9]{3}-?[89ab][a-f0-9]{3}-?[a-f0-9]{12}\Z', re.I)
        sk_regex = re.compile('^sub-[a-f0-9]{8}-?[a-f0-9]{4}-?4[a-f0-9]{3}-?[89ab][a-f0-9]{3}-?[a-f0-9]{12}\Z', re.I)
        ak_regex = re.compile('^s-[a-f0-9]{8}-?[a-f0-9]{4}-?4[a-f0-9]{3}-?[89ab][a-f0-9]{3}-?[a-f0-9]{12}\Z', re.I)
        
        assert pk_regex.match(pubKey), 'Invalid format of publish key'
        assert sk_regex.match(subKey), 'Invalid format of subscribe key'
        assert ak_regex.match(secretKey), 'Invalid format of secret key'
                
        self.pubKey = pubKey
        self.subKey = subKey
        self.secretKey = secretKey
        
        if cipher is not None:
            assert isinstance(cipher, str), 'Cipher expected as string'
            assert len(cipher), 'Minimal length of cipher key is 16 symbols'
            
            if len(cipher) > 16: cipher = cipher[:15]            
            self.cipher = AESCipher(cipher); 
        else:
            self.cipher = None
        
        self.ssl = True
        
        self.__consumerMap = {}
        self.__producerMap = {}
        self.__permissions = None
               
        self.__connectionSettings()            
        self.permissions();
        
    def setHost(self, host):
        """
        Sets host name of coordinating node [dedicated deployments]
        :param host: string URL of the host
        """    
        assert host is not None, 'Host name required'
        assert host is not None, 'Host name required'
        self.__host = host;
        
    
    def __doCerts(self, sts, sks):
        
        home = expanduser("~")
        
#         if os.path.exists(home + '/magistral') == False:
#             os.makedirs(home + '/magistral')
             
        if os.path.exists(home + '/magistral/tmp') == False:
            os.makedirs(home + '/magistral/tmp')
        
        with open(home + '/magistral/tmp/ts', 'wb') as f:
            f.seek(0)
            f.write(bytearray(base64.standard_b64decode(sts)))
            f.close()
    
        with open(home + '/magistral/tmp/ks', 'wb') as f:
            f.seek(0)
            f.write(bytearray(base64.standard_b64decode(sks)))                    
            f.close()           
                           
        ks = jks.KeyStore.load(home + '/magistral/tmp/ks', 'magistral')
            
        self.uid = JksHandler.writePkAndCerts(ks)
        
        if os.path.exists(home + '/magistral/tmp'): 
            shutil.rmtree(home + '/magistral/tmp')
        
    def __connectionSettings(self):
        
        url = "https://" + self.__host + "/api/magistral/net/connectionPoints";
        user = self.pubKey + "|" + self.subKey;
        
        home = expanduser("~")
#         
        def conPointsCallback(json, err):
            if (err != None):
                self.logger.error(err)
                return
            else:            
                self.logger.debug("Received Connection Points : %s", json)    
                self.settings = JsonConverter.connectionSettings(json)
                
                self.__doCerts(self.settings['ts'], self.settings['ks'])
                            
                for setting in (self.settings["pub"]["ssl"] if self.ssl else self.settings["pub"]["plain"]):        
                    p = KafkaProducer(bootstrap_servers = setting["bootstrap_servers"],
                                    compression_type = setting["compression_type"],
                                    value_serializer = setting["value_serializer"],
                                    key_serializer = setting["key_serializer"],
                                    security_protocol = 'SSL',
                                    ssl_check_hostname = False,
                                    ssl_keyfile = home + '/magistral/' + self.uid + '/key.pem',
                                    ssl_cafile = home + '/magistral/' + self.uid + '/ca.pem',
                                    ssl_certfile = home + '/magistral/' + self.uid + '/certificate.pem', 
                                    linger_ms = setting["linger_ms"],
                                    retries = setting["retries"],
                                    partitioner = None);  
                    
                    self.token = self.settings["meta"]["token"]
                    for key, val in setting.items():
                        p.config[key] = val
                    
                    p.config['client_id'] = self.token;          
                                                 
                    self.__producerMap[self.token] = p;                    
                    break;
                
                                               
                self.__initMqtt(self.token);
#         
        return RestApiManager.get(url, None, user, self.secretKey, lambda json, err: conPointsCallback(json, err));     
    
    def __initMqtt(self, token):
        self.logger.debug("Init MQTT with token : [%s]", token);
        
        clientId = "magistral.mqtt.gw." + token;
        username = self.pubKey + "|" + self.subKey;
        
        def conCallback(client, userdata, flags, rc):
            self.__mqtt.publish("presence/" + self.pubKey + "/" + token, payload=bytearray([0]), qos=1, retain=True);
        
        def messageReceivedCallback():
            pass
                
        self.__mqtt = mqtt.Client(clientId, True, None, mqtt.MQTTv311, transport="tls");        
        self.__mqtt.username_pw_set(username, self.secretKey);
        self.__mqtt.will_set(topic = "presence/" + self.pubKey + "/" + token, payload=bytearray([0]), qos=1, retain=True);
        
        self.__mqtt.on_connect = conCallback
        
        self.logger.debug("Connect to MQTT with token : [%s:%d]", self.__host, 1883);        
        self.__mqtt.connect(self.__host, port=1883, keepalive=60, bind_address="")
        

    def permissions(self, topic=None, callback=None):
        
        """ Retrieves user permissions

        :param topic: (optional) Topic name to get permissions for
        :param callback: (optional) callback with permissions -> array of :class:`PermMeta`.
        
        :return: :class: array of :class:`PermMeta`
        """
        
        if self.__permissions is None:
            url = "https://" + self.__host + "/api/magistral/net/permissions"
            
            params = None;
            if (topic != None): params = { "topic" : topic }
               
            auth = self.pubKey + "|" + self.subKey;   
            
            json = RestApiManager.get(url, params, auth, self.secretKey); # , lambda json, err: permsRestCallback(json, err)
            perms = JsonConverter.userPermissions(json);
            
            self.__permissions = [];
            self.__permissions.extend(perms);
           
            if (callback is not None): callback(self.__permissions);        
            return self.__permissions;
        else:
            if topic is None:
                if (callback is not None): callback(self.__permissions);        
                return self.__permissions;
            else:                
                for perm in self.__permissions:
                    if perm.topic() != topic : continue
                    
                    if (callback is not None): callback([perm]);        
                    return [perm];
                 
                if (callback is not None): callback(None);        
                return None;
                    

    def grant(self, user, topic, read, write, ttl=0, channel=-1, callback=None):
        
        """ Grants permanent or temporary permissions to access network resources
        
        :param user: (str) user name
        :param topic: (str) topic name
        :param read: (bool) permission to read from channel(s)
        :param write: (bool) permission to write to channel(s)
        
        :param ttl: (int) (optional) Time-to-Live for permission [0 - is permanent, > 0 - are temporary]
        :param channel: (int) (optional) Number of channel permission should be granted for [ -1 - to all in the topic, >=0 - number of specific channel]
       
        :param callback: (optional) callback with up-to-date permissions -> array of :class:`PermMeta`.
        
        :return: :class: array of :class:`PermMeta`
        """
        
        assert user is not None, 'User name is required'
        assert topic is not None, 'Topic is required'
        
        assert isinstance(read, bool) and isinstance(write, bool), 'read/write permissions must be type of bool'
        
        url = "https://" + self.__host + "/api/magistral/net/grant"
        params = { 'user': user, 'topic': topic, 'read': read, 'write': write }
        
        if (channel > -1): 
            params['channel'] = channel;
        
        if (ttl > 0):
            params["ttl"] = ttl;
       
        auth = self.pubKey + "|" + self.subKey;
        
        def updatedUserPermsCallback(_userPerms, err):
            perms = JsonConverter.userPermissions(_userPerms);
            if (callback != None): callback(perms, err);
            return perms;
        
        def grantRestCallback(json, err) :
            if (callback != None and err == None) : 
                url = "https://" + self.__host + "/api/magistral/net/user_permissions"
                RestApiManager.get(url, {"userName" : user}, auth, self.secretKey, lambda userPerms, err: updatedUserPermsCallback(userPerms, err))   
        
        RestApiManager.put(url, params, auth, self.secretKey, lambda json, err: grantRestCallback(json, err));


    def revoke(self, user, topic, channel=-1, callback=None):
        
        """ Revokes user permission
        
        :param user: (str) user name
        :param topic: (str) topic name

        :param channel: (int) (optional) Number of channel for revocation [ -1 - from all in the topic, >=0 - number of specific channel]
       
        :param callback: (optional) callback with up-to-date permissions -> array of :class:`PermMeta`.
        
        :return: :class: array of :class:`PermMeta`
        """
        
        assert user is not None, 'User name is required'
        assert topic is not None, 'Topic is required'
                
        url = "https://" + self.__host + "/api/magistral/net/revoke"
        params = { 'user': user, 'topic': topic }
        
        if (channel > -1): 
            params['channel'] = channel;
               
        auth = self.pubKey + "|" + self.subKey;
        
        def updatedUserPermsCallback(_userPerms, err):
            perms = JsonConverter.userPermissions(_userPerms);
            if (callback != None): callback(perms, err);
            return perms;
        
        def delRestCallback(json, err) :
            if (callback != None and err == None) : 
                url = "https://" + self.__host + "/api/magistral/net/user_permissions"
                RestApiManager.get(url, {"userName" : user}, auth, self.secretKey, lambda userPerms, err: updatedUserPermsCallback(userPerms, err))   
        
        RestApiManager.delete(url, params, auth, self.secretKey, lambda json, err: delRestCallback(json, err));


    def subscribe(self, topic, group="default", channel=-1, listener=None, callback=None):
        
        """ Subscribes to topic / channel
        
        :param topic: (str) topic name
        :param group: (str) group name
        
        :param channel: (int) (optional) Number of channel to subscribe [ -1 - to all in the topic, >=0 - number of specific channel]
       
        :param callback: (optional) callback with subscription ack -> :class:`SubMeta`.
        
        :return: :class: `SubMeta`
        """
        
        assert isinstance(topic, str), 'Topic must be type of str'
        assert isinstance(channel, int), 'Channel must be type of int'
        
        try :
            if group == None: group = "default"; 
            
            if group not in self.__consumerMap:
                self.__consumerMap[group] = {}
            
            cm = self.__consumerMap[group];

            consumersCount = 0;
            if self.ssl:
                consumersCount = len(self.settings["sub"]["ssl"]) if "ssl" in self.settings["sub"] else 0; 
            else :
                consumersCount = len(self.settings["sub"]["plain"]) if "plain" in self.settings["sub"] else 0;

            if len(cm) < consumersCount: # No enough consumers are there
                
                # TODO CIPHER
                
                for setting in (self.settings["sub"]["ssl"] if self.ssl else self.settings["sub"]["plain"]):
                    
                    bs = setting["bootstrap_servers"]
                    if (bs in cm): continue;
                         
                    threadId = 'thread_id_consumer_' + group
                    name = 'thread_name_consumer_' + group
                    
                    c = GroupConsumer(threadId, name, self.subKey, bs, group, self.__permissions, self.cipher, self.uid);
                    self.__consumerMap[group][bs] = c;
            
            for bs, gc in self.__consumerMap[group].items():
                
                def asgCallback(assignment, triggerCallbacks = False):
                    
                    meta = None;
                    
                    for asgm in assignment:
                        if (asgm[0] != self.subKey + "." + topic): continue;
                        
                        try:
                            meta = SubMeta(group, topic, asgm[1], bs);
                            if (triggerCallbacks == True and callback != None): callback(meta, None);
                            return meta;
                        except:
                            self.logger.error("ERROR = %s", sys.exc_info()[1]);
                            if (triggerCallbacks == True and callback != None): callback(None, sys.exc_info()[1]);
                            return None;
                        break;
                
                asgn = gc.subscribe(topic, channel, listener, lambda assignment : asgCallback(assignment, True))
                gc.start();
                
                res = asgCallback(asgn, False)                
                return res;
                 
        except:
            pass

    def unsubscribe(self, topic, channel=-1, callback=None): 
        
        """ Unsubscribes from topic / channel
        
        :param topic: (str) topic name
        
        :param channel: (int) (optional) Number of channel to unsubscribe from [ -1 - to all in the topic, >=0 - number of specific channel]
       
        :param callback: (optional) callback with unsubscription ack -> :class:`SubMeta`.
        
        :return: :class: `SubMeta`
        """
               
        for groupName, consmap in self.__consumerMap.items():
            for conString, gc in consmap.items():
                gc.unsubscribe(self.subKey + "." + topic);
                
                meta = SubMeta(groupName, topic, channel, conString);
                if (callback != None): callback(meta)
                return meta;
    
    def __recordMetadata2PubMeta(self, meta):
        assert isinstance(meta, RecordMetadata);
        return PubMeta(meta[0], int(meta[1]), meta[4])

    def publish(self, topic, msg, channel=-1, callback=None):
        
        """ Send messages to topic / channel(s)
        
        :param topic: (str) topic name
        :param msg: (bytes) message content
        
        :param channel: (int) (optional) Number of channel to send message to [ -1 - to all in the topic, >=0 - number of specific channel]
       
        :param callback: (optional) callback with publish ack -> :class:`PubMeta`.
        
        :return: :class: `PubMeta`
        """
        
        assert(topic is not None), 'Topic is required'
        assert(msg is not None and isinstance(msg, bytes)), 'Message body is required an an non-empty bytearray'
        
        if self.cipher is not None: # encrypt
            msg = self.cipher.encrypt(msg); # AES/ECB + Base64 encrypted string 
        
        try:            
            if topic == None:
                raise MagistralException("Topic name must be specified");
            
            topicMeta = self.topic(topic);
             
            if topicMeta == None:
                raise MagistralException("Topic " + topic + " cannot be found");
            
            if channel == None or channel < -1:
                channel = -1;
            
            chs = topicMeta.channels();
            if (channel not in chs):
                raise MagistralException("There is no channel [" + str(channel) + "] available for topic " + topic);
            
            if self.__producerMap == None or len(self.__producerMap) == 0:
                raise MagistralException("Unable to publish message -> Client is not connected to the Service");
            
            token = list(self.__producerMap.keys())[0];
            p = self.__producerMap[token];
            
            realTopic = self.pubKey + "." + topic;     
            
            key = bytes(self.secretKey + "-" + token, 'utf8')
            
            if channel == -1:
                for ch in chs:
                    p.send(topic = realTopic, value = msg, key = key, partition = int(ch));
            else: 
                future = p.send(topic = realTopic, value = msg, key = key, partition = int(channel)).add_callback(lambda ack : pubCallback(ack));
                        
                def pubCallback(ack):                    
                    if callback is not None: 
                        callback(self.__recordMetadata2PubMeta(ack))
                    
                ack = future.get(5);
                return self.__recordMetadata2PubMeta(ack);
            
        except:
            ex = list(sys.exc_info())
            self.logger.error("Error [%s] : %s", ex[0], ex[1])            
            raise MagistralException(ex[1]);

    def topics(self, callback=None):  
        
        """ Returns information about available topics / channels for user
        
        :return: array of :class:`TopicMeta`
        """
                  
        perms = self.permissions();
        
        metaList = [];            
        for pm in perms:            
            meta = TopicMeta(pm.topic(), list(pm.channels()))            
            metaList.append(meta);
            
        if callback != None: callback(metaList);            
        return metaList;
     

    def topic(self, topic, callback = None):
        
        """ Returns information about topic available for user
        
        :param topic: (str) topic name
         
        :return: array of :class:`TopicMeta`
        """
        
        assert topic is not None, 'Topic name required'
        
        perms = self.permissions(topic);
        if perms is None: return None
                    
        metaList = None;            
        for pm in perms: 
            metaList = TopicMeta(pm.topic(), pm.channels());
                        
        if callback is not None: callback(metaList);                      
        return metaList;
                
    def history(self, topic, channel, count, start = -1, callback=None):
        
        """ Returns last messages sent over channel 
        
        :param topic: (str) topic name
        :param channel: (int) channel number
        :param count: (int) quantity of messages to return
        
        :param start: (optional) start point of epoch to fetch messages from
        :param callback: (optional) callback with array of :class:`Message`
         
        :return: array of :class:`Message`
        """
        
        assert topic is not None, 'Topic name required'
        assert channel is not None and isinstance(channel, int), 'Channel number required as int parameter'
        
        assert count is not None, 'Number of records to return must be positive'
        
        bs = self.settings['sub']['plain'][0]['bootstrap_servers'];
        
        mc = MagistralConsumer(self.pubKey, self.subKey, self.secretKey, bs, None);
        
        res = []
        if start < 0:
            res.extend(mc.history(topic, channel, count));
        else:
            res.extend(mc.historyForTimePeriod(topic, channel, start, end = int(round(time.time() * 1000)), limit = count));
            
        if callback is not None: callable(res);
        return res;        
            
    def historyIn(self, topic, channel, start=0, end=int(round(time.time() * 1000)), callback=None):
        
        """ Returns sent messages within time interval 
        
        :param topic: (str) topic name
        :param channel: (int) channel number
       
        :param start: (optional) start of epoch
        :param end: (optional) end of epoch

        :param callback: (optional) callback with array of :class:`Message`
         
        :return: array of :class:`Message`
        """
        
        assert topic is not None, 'Topic name required'
        assert channel is not None and isinstance(channel, int), 'Channel number required as int parameter'
        
        bs = self.settings['sub']['plain'][0]['bootstrap_servers'];
        
        mc = MagistralConsumer(self.pubKey, self.subKey, self.secretKey, bs, None);
        res = mc.historyForTimePeriod(topic, channel, start, end)
        
        if res is not None:
            if callback is not None: callable(res);
        else: 
            return None;

    def close(self):
        
        """
        Terminates all network activity.
        Magistral should be re-instantiated after this
        """
        
        for bsmap in self.__consumerMap.values():
            for c in bsmap.values(): c.close();  
                  
        self.__mqtt.disconnect();
        
    logging.getLogger('kafka.conn').setLevel(logging.INFO)  
    logging.getLogger('kafka.client').setLevel(logging.INFO) 
    logging.getLogger('kafka.coordinator').setLevel(logging.INFO) 
    logging.getLogger('kafka.metrics.metrics').setLevel(logging.INFO) 
    logging.getLogger('kafka.producer.kafka').setLevel(logging.INFO)
    logging.getLogger('kafka.producer.sender').setLevel(logging.INFO)